@@ -7,7 +7,7 @@ import os
 import fnmatch
 from pathlib import Path
 from pyutilb.util import *
-from pyutilb import log, ocr_youdao
+from pyutilb import log, YamlBoot, BreakException, ocr_youdao
 import ast
 from SeleniumBoot.validator import Validator
 from SeleniumBoot.extractor import Extractor
@@ -23,31 +23,16 @@ from seleniumrequests.request import RequestsSessionMixin
 class MyWebDriver(RequestsSessionMixin, Chrome):
     pass
 
-# 跳出循环的异常
-class BreakException(Exception):
-    def __init__(self, condition):
-        self.condition = condition # 跳转条件
 
 # selenium基于yaml的启动器
-class Boot(object):
+class Boot(YamlBoot):
 
-    def __init__(self, driver: MyWebDriver):
-        # webdriver
-        self.driver = driver
-        # 步骤文件所在的目录
-        self.step_dir = None
-        # 已下载过的url对应的文件，key是url，value是文件
-        self.downloaded_files = {}
-        # 基础url
-        self._base_url = None
-        # 当前页面的校验器
-        self.validator = Validator(self.driver)
-        # 当前页面的提取器
-        self.extractor = Extractor(self.driver)
+    def __init__(self):
+        super().__init__()
         # 动作映射函数
-        self.actions = {
-            'sleep': self.sleep,
-            'print': self.print,
+        actions = {
+            'auto_close': self.auto_close,
+            'close_driver': self.close_driver,
             'base_url': self.base_url,
             'goto': self.goto,
             'get': self.get,
@@ -82,15 +67,9 @@ class Boot(object):
             'refresh': self.refresh,
             'forward': self.forward,
             'back': self.back,
-            'for': self.do_for,
-            'once': self.once,
-            'break_if': self.break_if,
-            'moveon_if': self.moveon_if,
             'moveon_if_exist_by': self.moveon_if_exist_by,
+            'break_if_exist_by': self.break_if_exist_by,
             'break_if_not_exist_by': self.break_if_not_exist_by,
-            'include': self.include,
-            'set_vars': self.set_vars,
-            'print_vars': self.print_vars,
             'validate_by_jsonpath': self.validate_by_jsonpath,
             'validate_by_xpath': self.validate_by_xpath,
             'validate_by_css': self.validate_by_css,
@@ -99,129 +78,82 @@ class Boot(object):
             'extract_by_css': self.extract_by_css,
             'extract_by_eval': self.extract_by_eval,
         }
-        set_var('boot', self)
-        # 当前文件
-        self.step_file = None
+        self.add_actions(actions)
 
-    '''
-    执行入口
-    :param step_files 步骤配置文件或目录的列表
-    '''
-    def run(self, step_files):
-        for path in step_files:
-            # 1 模式文件
-            if '*' in path:
-                dir, pattern = path.rsplit(os.sep, 1)  # 从后面分割，分割为目录+模式
-                if not os.path.exists(dir):
-                    raise Exception(f'步骤配置目录不存在: {dir}')
-                self.run_1dir(dir, pattern)
-                return
+        # webdriver
+        self.driver = None
+        # 已下载过的url对应的文件，key是url，value是文件
+        self.downloaded_files = {}
+        # 基础url
+        self._base_url = None
+        # 自动关闭选项
+        self.auto_close_option = {}
 
-            # 2 不存在
-            if not os.path.exists(path):
-                raise Exception(f'步骤配置文件或目录不存在: {path}')
-
-            # 3 目录: 遍历执行子文件
-            if os.path.isdir(path):
-                self.run_1dir(path)
-                return
-
-            # 4 纯文件
-            self.run_1file(path)
-
-    # 执行单个步骤目录: 遍历执行子文件
-    # :param path 目录
-    # :param pattern 文件名模式
-    def run_1dir(self, dir, pattern ='*.yml'):
-        # 遍历目录: https://blog.csdn.net/allway2/article/details/124176562
-        files = os.listdir(dir)
-        files.sort() # 按文件名排序
-        for file in files:
-            if fnmatch.fnmatch(file, pattern): # 匹配文件名模式
-                file = os.path.join(dir, file)
-                if os.path.isfile(file):
-                    self.run_1file(file)
-
-    # 执行单个步骤文件
-    # :param step_file 步骤配置文件路径
-    # :param include 是否inlude动作触发
-    def run_1file(self, step_file, include = False):
-        # 获得步骤文件的绝对路径
-        if include: # 补上绝对路径
-            if not os.path.isabs(step_file):
-                step_file = self.step_dir + os.sep + step_file
-        else: # 记录目录
-            step_file = os.path.abspath(step_file)
-            self.step_dir = os.path.dirname(step_file)
-
-        log.debug(f"加载并执行步骤文件: {step_file}")
-        # 获得步骤
-        steps = read_yaml(step_file)
-        self.step_file = step_file
-        # 执行多个步骤
-        self.run_steps(steps)
+    def run_action(self, action, param):
+        return super().run_action(action, param)
 
     # 执行多个步骤
     def run_steps(self, steps):
-        # 逐个步骤调用多个动作
-        for step in steps:
-            for action, param in step.items():
-                self.run_action(action, param)
+        # 延迟初始化driver
+        if self.driver == None:
+            self.init_driver()
 
-    '''
-    执行单个动作：就是调用动作名对应的函数
-    :param action 动作名
-    :param param 参数
-    '''
-    def run_action(self, action, param):
-        if 'for(' in action:
-            n = int(action[4:-1])
-            self.do_for(param, n)
-            return
-
-        if action not in self.actions:
-            raise Exception(f'无效动作: [{action}]')
-
-        # 调用动作对应的函数
-        log.debug(f"处理动作: {action}={param}")
-        func = self.actions[action]
-        func(param)
+        # 调用父类实现
+        super().run_steps(steps)
 
     # --------- 动作处理的函数 --------
-    # for循环
-    # :param steps 每个迭代中要执行的步骤
-    # :param n 循环次数
-    def do_for(self, steps, n = None):
-        label = f"for({n})"
-        if n == None:
-            n = sys.maxsize # 最大int，等于无限循环次数
-            label = f"for(∞)"
-        log.debug(f"-- 开始循环: {label} -- ")
-        try:
-            for i in range(n):
-                # i+1表示迭代次数比较容易理解
-                log.debug(f"第{i+1}次迭代")
-                set_var('for_i', i+1)
-                self.run_steps(steps)
-        except BreakException as e:  # 跳出循环
-            log.debug(f"-- 跳出循环: {label}, 跳出条件: {e.condition} -- ")
-        else:
-            log.debug(f"-- 结束循环: {label} -- ")
+    # 初始化driver
+    def init_driver(self, _ = None):
+        if self.driver != None:
+            return
 
-    # 执行一次子步骤，相当于 for(1)
-    def once(self, steps):
-        self.do_for(steps, 1)
+        # 浏览器驱动
+        option = ChromeOptions()
+        option.add_experimental_option('excludeSwitches', ['enable-automation'])
+        # 模拟手机设备 https://blog.csdn.net/qq_38316655/article/details/113739442
+        # option.add_experimental_option("mobileEmulation", {"deviceName": "Nexus 5"})
+        # self.driver = Chrome(options=option)
+        self.driver = MyWebDriver(options=option)
+        # 当前页面的校验器, 依赖于driver
+        self.validator = Validator(self.driver)
+        # 当前页面的提取器, 依赖于driver
+        self.extractor = Extractor(self.driver)
 
-    # 检查并继续for循环
-    def moveon_if(self, expr):
-        # break_if(条件取反)
-        self.break_if(f"not ({expr})")
+    # 关闭driver(浏览器)
+    def close_driver(self, _ = None):
+        if self.driver != None:
+            self.driver.quit()
+            self.driver = None
 
-    # 跳出for循环
-    def break_if(self, expr):
-        val = eval(expr, globals(), bvars)  # 丢失本地与全局变量, 如引用不了json模块
-        if bool(val):
-            raise BreakException(expr)
+    # 设置自动关闭
+    def auto_close(self, option):
+        if option != None:
+            self.auto_close_option = option
+
+    # 是否在完成时关闭浏览器
+    @property
+    def close_on_finish(self):
+        option = self.auto_close_option
+        if 'on_finish' in option:
+            return bool(option['on_finish'])
+
+        return False
+
+    # 是否在异常时关闭浏览器
+    @property
+    def close_on_exception(self):
+        option = self.auto_close_option
+        if 'on_exception' in option:
+            return bool(option['on_exception'])
+
+        return False
+
+    # 当前url
+    @property
+    def current_url(self):
+        if self.driver == None:
+            return None
+        return self.driver.current_url
 
     # 检查并继续for循环
     def moveon_if_exist_by(self, config):
@@ -232,29 +164,10 @@ class Boot(object):
         if not self.exist_by_any(config):
             raise BreakException(config)
 
-    # 加载并执行其他步骤文件
-    def include(self, step_file):
-        self.run_1file(step_file, True)
-
-    # 设置变量
-    def set_vars(self, vars):
-        for k, v in vars.items():
-            v = replace_var(v)  # 替换变量
-            set_var(k, v)
-
-    # 打印变量
-    def print_vars(self, _):
-        log.info(f"打印变量: {bvars}")
-
-    # 睡眠
-    def sleep(self, seconds):
-        seconds = replace_var(seconds)  # 替换变量
-        time.sleep(int(seconds))
-
-    # 打印
-    def print(self, msg):
-        msg = replace_var(msg)  # 替换变量
-        log.debug(msg)
+    # 跳出for循环
+    def break_if_exist_by(self, config):
+        if self.exist_by_any(config):
+            raise BreakException(config)
 
     # 解析响应
     def _analyze_response(self, res, config):
@@ -284,7 +197,7 @@ class Boot(object):
     # :param config {url, validate_by_jsonpath, validate_by_css, validate_by_xpath, extract_by_jsonpath, extract_by_css, extract_by_xpath, extract_by_eval}
     def goto(self, config = {}):
         url = self._get_url(config)
-        driver.get(url)
+        self.driver.get(url)
         # fix bug: 如果跳到table页，会异步加载，必须等加载完才能解析table，因此等等
         time.sleep(2)
 
@@ -374,12 +287,15 @@ class Boot(object):
                     newname = f"{save_file}-{i}"
                 if not os.path.exists(newname):
                     return newname
-            raise Exception('目录太多文件，建议新建目录')
+            raise Exception('Too many file in save_dir, please change other directory.')
 
         return save_file
 
     # 执行下载文件
     def _do_download(self, url, save_file):
+        # 忽略base64编码的图片, 如 data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7
+        if url.startswith('data:'):
+            return
         if url in self.downloaded_files:
             return self.downloaded_files[url]
 
@@ -390,7 +306,7 @@ class Boot(object):
         # 设置变量
         set_var('download_file', save_file)
         self.downloaded_files[url] = save_file
-        log.debug(f"下载文件: url为{url}, 另存为{save_file}")
+        log.debug(f"Dowload file: url is {url}, save path is{save_file}")
         return save_file
 
     # 从图片标签中下载图片
@@ -447,7 +363,7 @@ class Boot(object):
         captcha = ocr_youdao.recognize_text(file_path)
         # 设置变量
         set_var('captcha', captcha)
-        log.debug(f"识别验证码: 图片为{file_path}, 验证码为{captcha}")
+        log.debug(f"Recognize captcha: image file is {file_path}, captcha is {captcha}")
         # 删除文件
         #os.remove(file)
 
@@ -482,14 +398,13 @@ class Boot(object):
     # :param input_data 表单数据, key是输入框的路径, value是填入的值
     def _input_by_type_and_data(self, type, input_data):
         for name, value in input_data.items():
-            log.debug(f"替换变量： {name} = {value}")
             value = replace_var(value)  # 替换变量
 
             # 找到输入框
             try:
                 ele = self.find_by(type, name)
             except Exception as ex:  # 找不到元素
-                log.error(f"找不到输入元素{name}", exc_info = ex)
+                log.error(f"Input element not found{name}", exc_info = ex)
                 continue
 
             if ele.tag_name == 'select': # 设置输入框
@@ -503,9 +418,24 @@ class Boot(object):
                 js = f"arguments[0].value = '{value}'" # 原生js
                 self.driver.execute_script(js, ele)
             else:
+                style = ele.get_attribute('style')
+                # 开放隐藏的元素
+                mat = re.search(r'display:\s*none;', style)
+                if mat != None:
+                    style = style.replace(mat.group(), '')
+                    #ele.set_attribute('style', style) # 无此方法
+                    self.set_attribute(ele, 'style', style)
                 ele.clear() # 先清空
                 ele.send_keys(value) # 后输入
 
+    # js修改元素的属性值
+    def set_attribute(self, ele, name, value):
+        self.driver.execute_script("arguments[0].setAttribute(arguments[1],arguments[2])", ele, name, value)
+
+    # js删除元素的属性值
+    def remove_attribute(self, ele, name):
+        self.driver.execute_script("arguments[0].removeAttribute(arguments[1])", ele, name)
+
     # 根据指定类型，查找元素
     def find_by(self, type, path):
         return self.driver.find_element(type2by(type), path)
@@ -519,7 +449,7 @@ class Boot(object):
                 if type == 'xpath': # xpath支持变量
                     path = replace_var(path)
                 return self.driver.find_element(type2by(type), path)
-        raise Exception(f"没有查找类型: {config}")
+        raise Exception(f"Invalid find type: {config}")
 
     # 根据任一类型，查找元素
     def find_all_by_any(self, config):
@@ -528,7 +458,7 @@ class Boot(object):
             if type in config:
                 path = config[type]
                 return self.driver.find_elements(type2by(type), path)
-        raise Exception(f"没有查找类型: {config}")
+        raise Exception(f"Invalid find type: {config}")
 
     # 根据指定类型，检查元素是否存在
     def exist_by(self, type, path):
@@ -546,15 +476,6 @@ class Boot(object):
         except NoSuchElementException:
             return False
 
-    # 根据指定类型，查找元素的文本
-    def get_text_by(self, type, path):
-        ele = self.find_by(type, path)
-        return ele.text
-
-    # 根据指定类型，检查元素的文本是否等于
-    def check_text_by(self, type, path, txt):
-        return self.get_text_by(type, path) == txt
-
     # 点击按钮
     # :param config {css, xpath}
     def click_by(self, config):
@@ -603,7 +524,13 @@ class Boot(object):
     # 切换进入iframe
     # :param config {css, xpath}
     def switch_to_frame_by(self, config):
-        ele = self.find_by_any(config)
+        if config == None: # 默认进入最后一个iframe
+            try:
+                ele = self.find_by('css', 'iframe:last-child') # 只有一个iframe时报错：NoSuchElementException
+            except NoSuchElementException:
+                ele = self.find_by('css', 'iframe')
+        else:
+            ele = self.find_by_any(config)
         self.driver.switch_to.frame(ele)
 
     # 跳回到主框架页
@@ -719,28 +646,27 @@ class Boot(object):
 
 # cli入口
 def main():
-    global driver
-    # 浏览器驱动
-    option = ChromeOptions()
-    option.add_experimental_option('excludeSwitches', ['enable-automation'])
-    # driver = Chrome(options=option)
-    driver = MyWebDriver(options=option)
     # 基于yaml的执行器
-    boot = Boot(driver)
+    boot = Boot()
+    # 读元数据：author/version/description
+    dir = os.path.dirname(__file__)
+    meta = read_init_file_meta(dir + os.sep + '__init__.py')
+    # 步骤配置的yaml
+    step_files = parse_cmd('SeleniumBoot', meta['version'])
+    if len(step_files) == 0:
+        raise Exception("Miss step config file or directory")
     try:
-        # 步骤配置的yaml
-        if len(sys.argv) > 1:
-            step_files = sys.argv[1:]
-        else:
-            raise Exception("未指定步骤配置文件或目录")
         # 执行yaml配置的步骤
         boot.run(step_files)
+        # 完成时关闭浏览器
+        if boot.close_on_finish:
+            boot.close_driver()
     except Exception as ex:
-        log.error(f"异常环境:当前步骤文件为 {boot.step_file}, 当前url为 {driver.current_url}", exc_info = ex)
+        log.error(f"Exception occurs: current step file is {boot.step_file}, 当前url为 {boot.current_url}", exc_info = ex)
+        # 异常时关闭浏览器
+        if boot.close_on_exception:
+            boot.close_driver()
         raise ex
-    finally:
-        driver.quit()
-
 
 if __name__ == '__main__':
     main()
\ No newline at end of file
